"""Analysis of flow cytometry-based mKeima assays in Python."""

import os.path
from typing import Optional, Union

import pandas as pd
import numpy as np

from .read import import_from_directory, import_facs_csv, infer_setup_from_filename
from .analyze import calculate_mkeima_score, summarize, summarize_outliers


__version__ = "0.5.1"



from collections import UserDict
import re
from typing import Iterable, Optional
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


class ColorWheelDict(UserDict):
    """Lookup dictionary that maps keys to hex colors by using a color wheel.

    When a key is not present the first color of the color wheel is added as the value,
    and the color is moved from the beginning to the end of the color wheel. If no list
    of colors is specified, a default list of ten colors is added to the color wheel.
    It is also possible to manually set key and color pairs by using the same syntax as
    for a regular dictionary.
    """

    def __init__(self, colors: Optional[list[str]] = None):
        """Initializes a ColorWheelDict.

        Args:
            colors: Optional, a list of hex colors used for the color wheel. By default
                a list with ten colors is used.
        """
        self.data = {}

        if colors is not None:
            self.colors = colors
        else:
            self.colors = [
                "#80b1d3",
                "#fdb462",
                "#8dd3c7",
                "#bebada",
                "#fb8072",
                "#b3de69",
                "#fccde5",
                "#d9d9d9",
                "#bc80bd",
                "#ccebc5",
            ]
        self._color_wheel = self.colors.copy()

    def _next_color(self) -> str:
        color = self._color_wheel.pop(0)
        self._color_wheel.append(color)
        return color

    def __setitem__(self, key, value):
        is_hexcolor = re.search(r"^#(?:[0-9a-fA-F]{3}){1,2}$", value)
        if is_hexcolor:
            self.data[key] = value
        else:
            raise ValueError(f"the specified value {value} is not a hexcolor.")

    def __getitem__(self, key):
        if key not in self.data:
            self.data[key] = self._next_color()
        return self.data[key]


def facet_channel_distribution(
    data: pd.DataFrame,
    channel: str,
    split_on: str,
    group_by: str = None,
    hue: str = None,
    density: bool = True,
    kind: str = "hist",
) -> list[(plt.Figure, list[plt.Axes])]:
    """Plots signal distributions of a FACS channel, one figure per split group.

    Args:
        data: Dataset for plotting.
        channel: Column in 'data' that will be plotted.
        split_on: Column used to group data into different figures.
        group_by: Optional, specifies a column that is used to group data into
            different subplots.
        hue: Optional, specifies a column that is used to group data within each
            subplot, generating one distribution per unique variable.
        density: Default True. If true, the bins of each distribution are normalized to
            the total number of counts and the bin width, so that the area under the
            histogram integrates to 1
        kind: "hist" or "kde", default "hist. Whether to plot a histogram or a gaussian
            kernel density estimate. Note that "kde" is considerably slower and might
            not always work.

    Returns:
        List of tuples, each containing a Matplotlib figure object and a list of
        matplotlib axes objects. Figures correspond to groups defined by 'split_on',
        and axes to groups defined by 'group_by'.
    """

    figures = []
    for group_name, group_data in data.groupby(split_on):
        fig, axes = channel_distribution(
            group_data,
            channel,
            group_by=group_by,
            hue=hue,
            kind=kind,
            density=density,
        )
        fig.suptitle(f"{split_on}: {group_name}", fontsize=12)
        fig.subplots_adjust(top=0.85)
        figures.append((fig, axes))
    xlims = [[], []]
    for fig, axes in figures:
        xlims[0].append(axes[0].get_xlim()[0])
        xlims[1].append(axes[0].get_xlim()[1])
    xlims = (min(xlims[0]), max(xlims[1]))
    [axes[0].set_xlim(xlims) for fig, axes in figures]
    return figures


def channel_distribution(
    data: pd.DataFrame,
    channel: str,
    group_by: str = None,
    hue: str = None,
    kind: str = "hist",
    density: bool = True,
) -> (plt.Figure, list[plt.Axes]):
    """Plot signal distributions of a FACS channel.

    Args:
        data: Dataset for plotting.
        channel: Column in 'data' that will be plotted.
        group_by: Optional, specifies a column that is used to group data into different
            subplots.
        hue: Optional, specifies a column that is used to group data within each
            subplot, generating one distribution per unique variable.
        density: Default True. If true, the bins of each distribution are normalized to
            the total number of counts and the bin width, so that the area under the
            histogram integrates to 1
        kind: "hist" or "kde", default "hist. Whether to plot a histogram or a gaussian
            kernel density estimate. Note that "kde" is considerably slower and might
            not always work.

    Returns:
        A matplotlib figure object and a list of matplotlib axes objects. Axes
        correspond to groups defined by 'sub_group'.
    """
    color_mapping = ColorWheelDict()
    if hue is not None:
        colors = {r: color_mapping[r] for r in data[hue].unique()}
    else:
        colors = color_mapping.colors[0]

    if group_by is not None:
        num_plots = len(data[group_by].unique())
        data_sets = data.groupby(group_by)
    else:
        num_plots = 1
        data_sets = [("", data)]
    ylabel = "Cell density" if density else "Cell number"

    bins = np.linspace(data[channel].min() * 0.98, data[channel].max() * 1.02, 50)

    plotheight = 2.5
    legendheight = 1.2
    figheight = plotheight + legendheight
    figwidth = num_plots * 3
    figsize = (figwidth, figheight)

    sns.set_style("whitegrid")
    fig, axes = plt.subplots(1, num_plots, figsize=figsize, sharex=True, sharey=True)
    axes = axes if isinstance(axes, Iterable) else (axes,)

    for pos, (group_name, group_data) in enumerate(data_sets):
        ax = axes[pos]
        if kind == "kde":
            sns.kdeplot(
                data=group_data,
                x=channel,
                hue=hue,
                legend=True,
                common_norm=True,
                palette=colors,
                fill=True,
                ax=ax,
            )
        elif kind == "hist":
            if hue is not None:
                plot_data = []
                label = []
                for replicate, rep_data in group_data.groupby(hue):
                    plot_data.append(rep_data[channel])
                    label.append(replicate)
                plot_color = [colors[replicate] for replicate in label]
            else:
                plot_data = group_data[channel]
                plot_color = colors
                label = None
            ax.hist(
                plot_data, density=density, histtype="step", bins=bins, color=plot_color
            )
            ax.hist(
                plot_data,
                density=density,
                histtype="stepfilled",
                bins=bins,
                label=label,
                color=plot_color,
                alpha=0.3,
                lw=0,
            )
            ax.set_xlabel(channel)
            if hue is not None:
                ax.legend(title=hue)
        else:
            sns.histplot(
                data=group_data,
                x=channel,
                hue=hue,
                bins=bins,
                legend=True,
                common_norm=True,
                element="step",
                palette=colors,
                ax=ax,
            )

        ax.grid(axis="x")
        ax.tick_params(axis="x", which="minor", bottom=True)
        ax.tick_params(axis="y", which="minor", bottom=True)
        ax.set_title(group_name)

    # Formats axis labels and gridlines
    axes[0].set_yticks([])
    axes[0].set_ylabel(ylabel)
    for ax_pos, ax in enumerate(axes):
        for spine in ["bottom", "left"]:
            ax.spines[spine].set_color("#000000")
            ax.spines[spine].set_linewidth(1)
        ax.grid(False, axis="y")
        ax.grid(axis="x", linestyle="dashed", linewidth=1, color="#cccccc")
    sns.despine(top=True, right=True)

    # Movse legend to the bottom of the plot
    fig.tight_layout()
    if hue is not None:
        for ax in axes:
            old_legend = ax.legend_
            legend_handles = old_legend.legendHandles
            legend_labels = [t.get_text() for t in old_legend.get_texts()]
            legend_title = old_legend.get_title().get_text()
            old_legend.remove()
        if len(axes) == 1:
            ncol = 2
        else:
            ncol = max([3, min([5, len(legend_labels)])])
        fig.legend(
            legend_handles,
            legend_labels,
            title=legend_title,
            bbox_to_anchor=(0.5, 0),
            ncol=ncol,
            loc="lower center",
        )
        figure_space_for_legend = legendheight / figheight
        fig.subplots_adjust(bottom=figure_space_for_legend)
    return fig, axes


def replicate_boxplots(
    data: pd.DataFrame,
    channel: str,
    group_by: str = "Condition",
    split_on: Optional[str] = None,
    order: Optional[list[str]] = None,
    showmeans: bool = False,
) -> (plt.Figure, list[plt.Axes]):
    """Box plots for each replicate and sub group, one figure per main group.

    Args:
        data: Dataset for plotting.
        channel: Column in 'data' for plotting
        group_by: Default "Condition". Specifies a column that is used together with
            the "Replicate" to group data. Each group is shown as a separate box plot.
        split_on: Optional, column used to split data into different subplots.
        order: Optional, allows specifying a list of sub groups and their order for
            plotting. Values must be present in the column specified by sub group.
        showmeans: Optional, if true means are displayed on top of boxplots.

    Returns:
        A matplotlib figure object and a list of matplotlib axes objects. Axes
        correspond to groups defined by 'split_on'.
    """
    if order is None:
        group_order = sorted(data[group_by].unique())
    else:
        group_order = order
    if split_on is None:
        num_plots = 1
        data_groups = [(None, data)]
    else:
        num_plots = len(data[split_on].unique())
        data_groups = data.groupby(split_on)

    replicate_order = sorted(data["Replicate"].unique())

    plotheight = 2.8
    legendheight = 0.6
    figheight = plotheight + legendheight
    figwidth = num_plots * len(group_order) + (0.15 * num_plots)
    figsize = (figwidth, figheight)

    meanprops = {
        "marker": "o",
        "markersize": "7",
        "markerfacecolor": "white",
        "markeredgecolor": "black",
    }
    sns.set_theme(style="whitegrid")
    fig, axes = plt.subplots(1, num_plots, figsize=figsize, sharex=True, sharey=True)
    axes = axes if isinstance(axes, Iterable) else (axes,)
    for pos, (group_name, group_data) in enumerate(data_groups):
        ax = axes[pos]
        sns.boxplot(
            data=group_data,
            x=group_by,
            y=channel,
            hue="Replicate",
            order=group_order,
            hue_order=replicate_order,
            ax=ax,
            showfliers=False,
            color="#808080",
            linewidth=1.25,
            showmeans=showmeans,
            meanprops=meanprops,
        )
        if pos != 0:
            ax.set_ylabel("")
        ax.legend(loc="upper left")
        ax.set_title(group_name)
        ax.set_xticklabels(ax.get_xticklabels(), rotation=0)

        for spine in ["bottom", "left"]:
            ax.spines[spine].set_color("#000000")
            ax.spines[spine].set_linewidth(1)
        ax.grid(False, axis="x")
        ax.grid(axis="y", linestyle="dashed", linewidth=1, color="#cccccc")
        sns.despine(top=True, right=True)
    fig.tight_layout()
    fig.subplots_adjust(wspace=0.15)
    return fig, axes


def replicate_summary(
    data: pd.DataFrame,
    on: str,
    group_by: str = "Condition",
    split_on: Optional[str] = None,
    order: Optional[list[str]] = None,
) -> (plt.Figure, list[plt.Axes]):
    """Plots experiment summary by showing mean values and individual replicate points.

    Generates one subplot per cell line. Plots for each condition the replicate mean
    as a box plot, and individual replicate data points as scatter points.

    Args:
        data: Dataset for plotting.
        on: Variable in 'data' for plotting.
        split_on: Optional, column used to split data into different subplots.
        group_by: Default "Condition". Specifies a column that is used to group data.
            Each group mean is shown as a separate bar plot, and individual data points
            are shown as scatter points.
        order: Optional, allows specifying a list of sub groups and their order for
            plotting. Values must be present in the column specified by sub group.

    Returns:
        Matplotlib figure object and a list of matplotlib axes objects.
    """
    if order is None:
        group_order = sorted(data[group_by].unique())
    else:
        group_order = order
    if split_on is None:
        num_plots = 1
        data_groups = [(None, data)]
    else:
        num_plots = len(data[split_on].unique())
        data_groups = data.groupby(split_on)

    plotheight = 2.5
    legendheight = 0.8
    figheight = plotheight + legendheight
    plotwidth = 0.5 * num_plots * len(group_order)
    spacingwidth = 0.75 * (num_plots - 1)
    figwidth = plotwidth + spacingwidth + 0.25
    figsize = (figwidth, figheight)

    color_mapping = ColorWheelDict()
    colors = {c: color_mapping[c] for c in group_order}
    sns.set_theme(style="whitegrid")
    fig, axes = plt.subplots(1, num_plots, figsize=figsize, sharex=True, sharey=True)
    axes = axes if isinstance(axes, Iterable) else (axes,)
    for pos, (group_name, group_data) in enumerate(data_groups):
        ax = axes[pos]
        sns.barplot(
            data=group_data,
            x=group_by,
            y=on,
            order=group_order,
            alpha=1,
            ci=None,
            edgecolor="#000000",
            linewidth=0.75,
            ax=ax,
            palette=colors,
        )
        sns.swarmplot(
            data=group_data,
            x=group_by,
            y=on,
            order=group_order,
            dodge=False,
            edgecolor="#000000",
            linewidth=0.75,
            ax=ax,
            palette=colors,
        )
        if pos != 0:
            ax.set_ylabel("")
        ax.set_title(group_name)
        ax.set_xticklabels(ax.get_xticklabels(), rotation=90)

        for spine in ["bottom", "left"]:
            ax.spines[spine].set_color("#000000")
            ax.spines[spine].set_linewidth(1)
        ax.grid(False, axis="x")
        ax.grid(axis="y", linestyle="dashed", linewidth=1, color="#cccccc")
        sns.despine(top=True, right=True)
    fig.tight_layout()
    fig.subplots_adjust(wspace=0.75)
    return fig, axes


def set_dpi(dpi: int) -> None:
    """Changes the default dots per inch settings for matplotlib plots.

    This effectively makes figures smaller or larger, without affecting the relative
    sizes of elements within the figures.

    Args:
        dpi: New default dots per inch.
    """
    plt.rcParams["figure.dpi"] = dpi



"""This module provides functions for analyzing FACS data.

The following functions are available:

- `calculate_mkeima_score`: Adds "low pH mkeima" to "high pH mkeima" scores. mKeima
    scores are log ratios of low pH to high pH mKeima signals, normalized to a reference
    sample by calculating z-scores using the median and standard deviation of the
    reference sample.
- `summarize`: Calculates summary statistics (mean, median, std, and number  of cells).
- `summarize_outliers`: Calculates the amount of outliers for each condition and
    replicate. Outliers are defined as events with a value above the a certain
    percentile (by default 98.5%) of the reference population.
"""
from typing import Optional

import pandas as pd
import numpy as np


def calculate_mkeima_score(
    data: pd.DataFrame,
    high_ph: str,
    low_ph: str,
    reference: str,
    group_by: Optional[str] = None,
    log_transform: bool = False,
) -> None:
    """Adds normalized "low pH mkeima" to "high pH mkeima" scores.

    Mkeima scores are normalized log ratios of low pH to high pH mKeima signals and
    calculated as follows. Low pH mKeima signals are divided by high pH mKeima signals
    and the ratios are log transformed. Log ratios are then normalized to the reference
    condition by subtracting the median of the of the reference population and then
    dividing by the standard deviation of the reference population. Normalized log
    ratios are added to the dataframe column "mkeima score".

    Args:
        data: Dataframe containing FACS data.
        high_ph: Column corresponding to the high pH mKeima channel.
        low_ph: Column corresponding to the low pH mKeima channel.
        reference: Reference condition used to normalize ratios of low pH to high pH
            ratios. Must correspond to a value in data["Condition"].
        group_by: Optional, if specified the dataframe is grouped by unique values of
            this column and each group is normalized independently to the reference
            condition.
        log_transform: Default False. If True low pH and high pH values are log
            transformed before calculating mkeima scores. Only use this option when the
            reported FACS intensities are not already in log-space. It is recommended to
            export channel values from FlowJo, which are already log transformed.
    """
    if reference not in data["Condition"].unique():
        raise KeyError(
            f'The reference condition "{reference}" is not '
            f'present in data["Condition"]!'
        )

    if log_transform:
        data["mkeima score"] = np.log(data[low_ph]) - np.log(data[high_ph])
    else:
        data["mkeima score"] = data[low_ph] - data[high_ph]

    group_masks = []
    if group_by is not None:
        for group_name in data[group_by].unique():
            group_masks.append(data[group_by] == group_name)
    else:
        group_masks.append(np.ones(data.shape[0], dtype=bool))

    for group_mask in group_masks:
        group_data = data.loc[group_mask]

        norm_data = group_data.loc[(group_data["Condition"] == reference)]
        norm_median = np.median(norm_data["mkeima score"])
        norm_std = np.std(norm_data["mkeima score"])

        mkeima_scores = (group_data["mkeima score"] - norm_median) / norm_std
        mkeima_scores = mkeima_scores
        data.loc[group_mask, "mkeima score"] = mkeima_scores


def summarize(
    data: pd.DataFrame,
    group_by: list[str] = ["Condition", "Replicate"],
    on: str = "mkeima score",
) -> pd.DataFrame:
    """Calculates summary statistics.

    Total number of cells (i.e. FACS events), mean, standard deviation, median.

    Args:
        data: Dataframe used for summarizing.
        group_by: Default ["Condition", "Replicate"]. List of dataframe columns that
            will be used for grouping data before summarizing each group individually.
            Should include all columns that are necessary to distinguish each unique
            sample.
        on: Default "mkeima score". Specifies column that will be used for summarizing.

    Returns:
        A dataframe containing the columns used for grouping, as well as "Median",
        "Mean", "Std", and "Total cells".
    """
    results = (
        data.groupby(by=group_by)
        .agg(
            Median=(on, np.median),
            Mean=(on, np.mean),
            Std=(on, np.std),
            Total_cells=(on, "size"),
        )
        .reset_index()
    )
    results.rename(columns={"Total_cells": "Total cells"}, inplace=True)
    return results


def summarize_outliers(
    data: pd.DataFrame,
    reference: str,
    on: str = "mkeima score",
    reference_percentile: float = 98.5,
    group_by: Optional[list] = None,
) -> pd.DataFrame:
    """Calculates the amount of outliers for each condition and replicate.

    The threshold for outliers is defined as the 98.5 percentile of the mkeima score
    distribution from the reference condition. Entries in the dataframe are grouped
    according to unique values in the "Condition" and "Replicate" columns, and for each
    group the number and relative amount of outliers is calculated.

    Args:
        data: Dataframe used for summarizing. Must contain the columns "Condition" and
        "Replicate", as well as columns specified by the 'on' and 'group_by' arguments.
        reference: Reference condition that is used for calculating an outlier
            threshold. Must correspond to a value in data["Condition"].
        on: Default "mkeima score". Specifies column that will be used for calculation
            of outliers.
        reference_percentile: Percentile of the reference condition population that is
            used to calculate the outlier threshold.
        group_by: Optional, list of dataframe columns. If specified, the dataframe is
            first grouped by unique values of these columns and the outlier threshold
            and outliers are calculated for each group separately.

    Returns:
        A dataframe containing the columns "Condition", "Replicate", "Total cells",
        "Outliers", "Outliers [%]", and the columns specified by the 'group_by'
        argument.
    """
    default_grouping = ["Condition", "Replicate"]
    if group_by is not None:
        group_by = [by for by in group_by if by not in default_grouping]
        if not group_by:
            group_by = None

    if group_by is None:
        grouping = [(None, data)]
    else:
        grouping = data.groupby(group_by)

    group_results = []
    for group_name, group_data in grouping:
        reference_mask = group_data["Condition"] == reference
        reference_scores = group_data.loc[reference_mask, on]
        cutoff_uppers = np.percentile(reference_scores, reference_percentile)

        results = (
            group_data.groupby(by=default_grouping)
            .agg(
                Total_cells=(on, "size"),
                Outliers=(on, lambda x: (x >= cutoff_uppers).sum()),
            )
            .reset_index()
        )
        results["Outliers [%]"] = (results["Outliers"] / results["Total_cells"]) * 100
        results.rename(columns={"Total_cells": "Total cells"}, inplace=True)

        # Add columns from the user specified grouping
        if group_name is not None:
            if isinstance(group_name, str):
                group_name = (group_name,)
            for column_value, column in zip(group_name, group_by):
                results[column] = column_value
        group_results.append(results)
    results_table = pd.concat(group_results, ignore_index=True)
    return results_table



import pathlib
from typing import Iterator, Optional, Union

import pandas as pd


def import_from_directory(
    root: str, levels: list[str] = ["Experiment"], file_extension: str = "csv"
) -> pd.DataFrame:
    """Reads FACS data files from a nested directory and returns a dataframe.

    Use an empty string for a level to prevent adding this level as a new column to the
    dataframe.

    The length of 'levels' corresponds to the depth of subdirectories that is walked
    down to reach the folder from which FACS data files are imported.

    Args:
        root: Root directory for importing FACS data files.
        levels: Default ["Experiment"]. A list of strings that represent the
            subdirectory structure of the specified root directory. The first entry
            corresponds to the root directory. Each level is added as a new column to
            the imported data and filled with the actual folder name.
        file_extension: File extension to identify FACS data files. Must not contain the
            dot that separates filename from file extension.

    Returns:
        A dataframe containing all imported data.
    """
    if not file_extension.startswith("."):
        file_extension = f".{file_extension}"

    target_subdir_depth = len(levels) - 1

    dataframes = []
    for path_entry, subdir_depth in _iterdir(root):
        if (
            subdir_depth == target_subdir_depth
            and path_entry.suffix.lower() == file_extension
        ):
            dataframe = import_facs_csv(path_entry)
            for pos, column in enumerate(levels[::-1]):
                if column:
                    dataframe[column] = path_entry.parents[pos].name
            dataframes.append(dataframe)

    if dataframes:
        data = pd.concat(dataframes)
    else:
        data = pd.DataFrame()
    return data


def import_facs_csv(filepath: str) -> pd.DataFrame:
    """Reads FACS data from a CSV file and returns a dataframe.

    Note that also also other value delimiters are allowed, such as tab, because the
    delimiter is automatically detected by the pandas.read_csv() function.

    Args:
        filepath: Path of CSV file containing FACS data.

    Returns:
        A dataframe containing FACS data. Also adds the column "Filename", containing
        the filename without the file extension.
    """
    filepath = pathlib.Path(filepath)
    data = pd.read_csv(filepath)
    data["Filename"] = filepath.stem
    return data


def infer_setup_from_filename(
    data: pd.DataFrame,
    strip: Union[list[str], str] = ["export_", "_Singlets"],
    inplace: bool = False,
) -> Optional[pd.DataFrame]:
    """Infers condition and replicate from filenames and adds them to the dataframe.

    To extract condition and replicate, entries from the "Filename" column are split at
    the last underscore, with the first part being used as the condition and the last
    part as the replicate. Inferred values are added as new columns "Condition" and
    "Replicate". Filename entries should not contain file extensions.

    Args:
        data: Dataframe containing a column "Filename", which is used for inferring the
            experimental condition and the replicate.
        strip_filename: String or list of strings that will be removed from the filename
            before replicate and condition are extracted.
        inplace: Default False. If True, performs operation inplace and returns None.

    Returns:
        A copy of the dataframe containing the additional columns "Condition" and
        "Replicate".
    """
    strip = [strip] if isinstance(strip, str) else strip
    filename_mapping = {f: f for f in data["Filename"].unique()}
    for string in strip:
        for filename, mapping in filename_mapping.items():
            filename_mapping[filename] = mapping.replace(string, "")
    replicate_mapping = {}
    condition_mapping = {}
    for filename, mapping in filename_mapping.items():
        replicate_mapping[filename] = mapping.split("_")[-1]
        condition_mapping[filename] = "_".join(mapping.split("_")[:-1])
    replicates = [replicate_mapping[filename] for filename in data["Filename"]]
    conditions = [condition_mapping[filename] for filename in data["Filename"]]

    if not inplace:
        data = data.copy()
        data["Replicate"] = replicates
        data["Condition"] = conditions
        return data
    else:
        data["Replicate"] = replicates
        data["Condition"] = conditions
        return None


def _iterdir(path: str, curr_subdir_depth: int = 0) -> Iterator[tuple[str, int]]:
    """Recursive iterator that returns filenames and directory depth of iteration."""
    for path_entry in pathlib.Path(path).iterdir():
        if path_entry.is_file():
            yield path_entry, curr_subdir_depth
        else:
            yield from _iterdir(path_entry, curr_subdir_depth=curr_subdir_depth + 1)
